from setuptools import setup
setup(
    name='coj_judger',
    version='0.17',
    description='A Framework for developing a judger of COJ.',
    author='Caiwen',
    author_email='3102733279@qq.com',
    url='https://github.com/CZOI-Project/JudgerFramework',
    packages=['coj_judger'],
)
